'use strict';

// Returns the width, height and offsets of the provided element
var getElementClientRect = (function (element) {
  // get the basic client rect, it doesn't include margins
  var width = element.offsetWidth;
  var height = element.offsetHeight;
  var y = element.offsetTop;
  var x = element.offsetLeft;
  return {
    width: width,
    height: height,
    y: y,
    x: x
  };
});

var getNodeName = (function (element) {
  return element ? element.nodeName.toUpperCase() : null;
});

var getParentNode = (function (element) {
  if (getNodeName(element) === 'HTML') {
    // DocumentElement detectedF
    return element;
  }

  return element.parentNode || // DOM Element detected
  // $FlowFixMe: need a better way to handle this...
  element.host || // ShadowRoot detected
  document.ownerDocument || // Fallback to ownerDocument if available
  document.documentElement // Or to documentElement if everything else fails
  ;
});

function getWindow(node) {
  var ownerDocument = node.ownerDocument;
  return ownerDocument ? ownerDocument.defaultView : window;
}

function getComputedStyle(element) {
  return getWindow(element).getComputedStyle(element);
}

function getScrollParent(node) {
  if (!node) {
    return document.body;
  }

  if (['HTML', 'BODY', '#document'].includes(node.nodeName.toUpperCase())) {
    return node.ownerDocument.body;
  }

  if (node instanceof HTMLElement) {
    // Firefox want us to check `-x` and `-y` variations as well
    var _getComputedStyle = getComputedStyle(node),
        overflow = _getComputedStyle.overflow,
        overflowX = _getComputedStyle.overflowX,
        overflowY = _getComputedStyle.overflowY;

    if (/(auto|scroll|overlay)/.test(overflow + overflowY + overflowX)) {
      return node;
    }
  }

  return getScrollParent(getParentNode(node));
}

function listScrollParents(element, list) {
  if (list === void 0) {
    list = [];
  }

  var scrollParent = getScrollParent(element);
  var isBody = getNodeName(scrollParent) === 'BODY';
  var target = isBody ? getWindow(scrollParent) : scrollParent;
  var updatedList = list.concat(target);
  return isBody ? updatedList : updatedList.concat(listScrollParents(getParentNode(target)));
}

var getElementMargins = (function (element) {
  // get the element margins, we need them to properly align the popper
  var styles = getComputedStyle(element);
  var top = parseFloat(styles.marginTop) || 0;
  var right = parseFloat(styles.marginRight) || 0;
  var bottom = parseFloat(styles.marginBottom) || 0;
  var left = parseFloat(styles.marginLeft) || 0;
  return {
    top: top,
    right: right,
    bottom: bottom,
    left: left
  };
});

var addClientRectMargins = (function (rect, element) {
  var margins = getElementMargins(element);
  return {
    width: rect.width + margins.left + margins.right,
    height: rect.height + margins.top + margins.bottom,
    y: rect.y - margins.top,
    x: rect.x - margins.left
  };
});

var unwrapJqueryElement = (function (element) {
  return (// $FlowFixMe: need to get type refinement work
    element.hasOwnProperty('jquery') ? element[0] : element
  );
});

// source: https://stackoverflow.com/questions/49875255
var order = function order(modifiers) {
  // indexed by name
  var mapped = modifiers.reduce(function (mem, i) {
    mem[i.name] = i;
    return mem;
  }, {}); // inherit all dependencies for a given name

  var inherited = function inherited(i) {
    return mapped[i] && mapped[i].requires ? mapped[i].requires.reduce(function (mem, i) {
      return [].concat(mem, [i], inherited(i));
    }, []) : [];
  }; // order ...


  var ordered = modifiers.sort(function (a, b) {
    return !!~inherited(b.name).indexOf(a.name) ? -1 : 1;
  });
  return ordered;
};

var orderModifiers = (function (modifiers) {
  return [].concat(order(modifiers.filter(function (_ref) {
    var phase = _ref.phase;
    return phase === 'read';
  })), order(modifiers.filter(function (_ref2) {
    var phase = _ref2.phase;
    return phase === 'main';
  })), order(modifiers.filter(function (_ref3) {
    var phase = _ref3.phase;
    return phase === 'afterMain';
  })), order(modifiers.filter(function (_ref4) {
    var phase = _ref4.phase;
    return phase === 'write';
  })));
});

function expandToHashMap(value, keys) {
  return keys.reduce(function (hashMap, key) {
    hashMap[key] = value;
    return hashMap;
  }, {});
}

function microtaskDebounce(fn) {
  var called = false;
  return function () {
    return new Promise(function (resolve) {
      if (called) {
        return resolve();
      }

      called = true;
      Promise.resolve().then(function () {
        called = false;
        resolve(fn());
      });
    });
  };
}

var format = (function (str) {
  for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    args[_key - 1] = arguments[_key];
  }

  return [].concat(args).reduce(function (p, c) {
    return p.replace(/%s/, c);
  }, str);
});

var top = 'top';
var bottom = 'bottom';
var right = 'right';
var left = 'left';
var basePlacements = [top, bottom, right, left];
var start = 'start';
var end = 'end';
var placements = basePlacements.reduce(function (acc, placement) {
  return acc.concat([placement, placement + "-" + start, placement + "-" + end]);
}, []); // modifiers that need to read the DOM

var read = 'read'; // pure-logic modifiers

var main = 'main'; // pure-logic modifiers that run after the main phase (such as computeStyles)

var write = 'write';

var ERROR_MESSAGE = 'Popper: modifier "%s" provided an invalid %s property, expected %s but got %s';
var VALID_PROPERTIES = ['name', 'enabled', 'phase', 'fn', 'onLoad', 'requires', 'options'];
var validateModifiers = (function (modifiers) {
  modifiers.forEach(function (modifier) {
    Object.keys(modifier).forEach(function (key) {
      switch (key) {
        case 'name':
          if (typeof modifier.name !== 'string') {
            console.error(format(ERROR_MESSAGE, String(modifier.name), '"name"', '"string"', "\"" + String(modifier.name) + "\""));
          }

          break;

        case 'enabled':
          if (typeof modifier.enabled !== 'boolean') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"enabled"', '"boolean"', "\"" + String(modifier.enabled) + "\""));
          }

        case 'phase':
          if (![read, main, write].includes(modifier.phase)) {
            console.error(format(ERROR_MESSAGE, modifier.name, '"phase"', 'either "read", "main" or "write"', "\"" + String(modifier.phase) + "\""));
          }

          break;

        case 'fn':
          if (typeof modifier.fn !== 'function') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"fn"', '"function"', "\"" + String(modifier.fn) + "\""));
          }

          break;

        case 'onLoad':
          if (typeof modifier.onLoad !== 'function') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"onLoad"', '"function"', "\"" + String(modifier.fn) + "\""));
          }

          break;

        case 'onDestroy':
          if (typeof modifier.onLoad !== 'function') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"onDestroy"', '"function"', "\"" + String(modifier.fn) + "\""));
          }

          break;

        case 'requires':
          if (!Array.isArray(modifier.requires)) {
            console.error(format(ERROR_MESSAGE, modifier.name, '"requires"', '"array"', "\"" + String(modifier.requires) + "\""));
          }

          break;

        case 'options':
        case 'data':
          break;

        default:
          console.error("PopperJS: an invalid property has been provided to the \"" + modifier.name + "\" modifier, valid properties are " + VALID_PROPERTIES.map(function (s) {
            return "\"" + s + "\"";
          }).join(', ') + "; but \"" + key + "\" was provided.");
      }
    });
  });
});

// DOM Utils
var INVALID_ELEMENT_ERROR = 'Popper: Invalid reference or popper argument provided to Popper, they must be either a valid DOM element or a jQuery-wrapped DOM element.';
var INFINITE_LOOP_ERROR = 'Popper: An infinite loop in the modifiers cycle has been detected! The cycle has been interrupted to prevent a browser crash.';

var areValidElements = function areValidElements() {
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  return !args.some(function (element) {
    return !(element instanceof Element);
  });
};

var defaultOptions = {
  placement: 'bottom',
  eventListeners: {
    scroll: true,
    resize: true
  },
  modifiers: [],
  strategy: 'absolute'
};

var Popper =
/*#__PURE__*/
function () {
  function Popper(reference, popper, options) {
    var _this = this;

    if (options === void 0) {
      options = defaultOptions;
    }

    this.state = {
      placement: 'bottom',
      orderedModifiers: [],
      options: defaultOptions,
      modifiersData: {}
    };
    this.update = microtaskDebounce(function () {
      return new Promise(function (success) {
        _this.forceUpdate();

        success(_this.state);
      });
    });
    // Unwrap `reference` and `popper` elements in case they are
    // wrapped by jQuery, otherwise consume them as is
    var referenceElement = unwrapJqueryElement(reference);
    var popperElement = unwrapJqueryElement(popper);
    this.state.elements = {
      reference: referenceElement,
      popper: popperElement
    }; // Store options into state

    this.state.options = Object.assign({}, defaultOptions, {}, options); // Don't proceed if `reference` or `popper` are invalid elements

    if (!areValidElements(referenceElement, popperElement)) {
      if (process.env.NODE_ENV !== "production") {
        console.error(INVALID_ELEMENT_ERROR);
      }

      return;
    }

    this.state.scrollParents = {
      reference: listScrollParents(referenceElement),
      popper: listScrollParents(popperElement)
    }; // Validate the provided modifiers so that the consumer will get warned
    // of one of the custom modifiers is invalid for any reason

    if (process.env.NODE_ENV !== "production") {
      validateModifiers(this.state.options.modifiers);
    } // Order `options.modifiers` so that the dependencies are fulfilled
    // once the modifiers are executed


    this.state.orderedModifiers = orderModifiers([].concat(this.constructor.defaultModifiers, this.state.options.modifiers)) // Apply user defined preferences to modifiers
    .map(function (modifier) {
      return Object.assign({}, modifier, {}, _this.state.options.modifiers.find(function (_ref) {
        var name = _ref.name;
        return name === modifier.name;
      }));
    }); // Modifiers have the opportunity to execute some arbitrary code before
    // the first update cycle is ran, the order of execution will be the same
    // defined by the modifier dependencies directive.
    // The `onLoad` function may add or alter the options of themselves

    this.state.orderedModifiers.forEach(function (_ref2) {
      var onLoad = _ref2.onLoad,
          enabled = _ref2.enabled;
      return enabled && onLoad && onLoad(_this.state);
    });
    this.update().then(function () {
      // After the first update completed, enable the event listeners
      _this.enableEventListeners(_this.state.options.eventListeners);
    });
  } // Async and optimistically optimized update
  // it will not be executed if not necessary
  // debounced, so that it only runs at most once-per-tick


  var _proto = Popper.prototype;

  // Syncronous and forcefully executed update
  // it will always be executed even if not necessary, usually NOT needed
  // use Popper#update instead
  _proto.forceUpdate = function forceUpdate() {
    var _this2 = this;

    var _this$state$elements = this.state.elements,
        referenceElement = _this$state$elements.reference,
        popperElement = _this$state$elements.popper; // Don't proceed if `reference` or `popper` are not valid elements anymore

    if (!areValidElements(referenceElement, popperElement)) {
      if (process.env.NODE_ENV !== "production") {
        console.error(INVALID_ELEMENT_ERROR);
      }

      return;
    } // Get initial measurements
    // these are going to be used to compute the initial popper offsets
    // and as cache for any modifier that needs them later


    this.state.measures = {
      reference: getElementClientRect(referenceElement),
      // CSS marginsc an be applied to popper elements to quickly
      // apply offsets dynamically based on some CSS selectors.
      // For this reason we include margins in this calculation.
      popper: addClientRectMargins(getElementClientRect(popperElement), popperElement)
    }; // Modifiers have the ability to read the current Popper state, included
    // the popper offsets, and modify it to address specifc cases

    this.state.reset = false; // Cache the placement in cache to make it available to the modifiers
    // modifiers will modify this one (rather than the one in options)

    this.state.placement = this.state.options.placement;
    this.state.orderedModifiers.forEach(function (modifier) {
      return _this2.state.modifiersData[modifier.name] = Object.assign({}, modifier.data);
    });
    var __debug_loops__ = 0;

    for (var index = 0; index < this.state.orderedModifiers.length; index++) {
      if (process.env.NODE_ENV !== "production") {
        __debug_loops__ += 1;

        if (__debug_loops__ > 100) {
          console.error(INFINITE_LOOP_ERROR);
          break;
        }
      }

      if (this.state.reset === true) {
        this.state.reset = false;
        index = -1;
        continue;
      }

      var _this$state$orderedMo = this.state.orderedModifiers[index],
          fn = _this$state$orderedMo.fn,
          enabled = _this$state$orderedMo.enabled,
          options = _this$state$orderedMo.options;

      if (enabled && typeof fn === 'function') {
        this.state = fn(this.state, options);
      }
    }
  };

  _proto.enableEventListeners = function enableEventListeners(eventListeners) {
    var _this3 = this;

    var _ref3 = typeof eventListeners === 'boolean' ? expandToHashMap(eventListeners, ['scroll', 'resize']) : eventListeners,
        scroll = _ref3.scroll,
        resize = _ref3.resize;

    if (scroll) {
      var scrollParents = [].concat(this.state.scrollParents.reference, this.state.scrollParents.popper);
      scrollParents.forEach(function (scrollParent) {
        return scrollParent.addEventListener('scroll', _this3.update, {
          passive: true
        });
      });
    }

    if (resize) {
      var window = getWindow(this.state.elements.popper);
      window.addEventListener('resize', this.update, {
        passive: true
      });
    }
  };

  _proto.destroy = function destroy() {
    var _this4 = this;

    // Remove scroll event listeners
    var scrollParents = [].concat(this.state.scrollParents.reference, this.state.scrollParents.popper);
    scrollParents.forEach(function (scrollParent) {
      return scrollParent.removeEventListener('scroll', _this4.update);
    }); // Remove resize event listeners

    var window = getWindow(this.state.elements.popper);
    window.removeEventListener('resize', this.update); // Run `onDestroy` modifier methods

    this.state.orderedModifiers.forEach(function (_ref4) {
      var onDestroy = _ref4.onDestroy,
          enabled = _ref4.enabled;
      return enabled && onDestroy && onDestroy(_this4.state);
    });
  };

  return Popper;
}();
Popper.defaultModifiers = [];

var getBasePlacement = (function (placement) {
  return placement.split('-')[0];
});

var getVariationPlacement = (function (placement) {
  return placement.split('-')[1];
});

var getMainAxisFromPlacement = (function (placement) {
  return ['top', 'bottom'].includes(placement) ? 'x' : 'y';
});

var computeOffsets = (function (_ref) {
  var reference = _ref.reference,
      element = _ref.element,
      placement = _ref.placement,
      scroll = _ref.scroll;
  var basePlacement = placement ? getBasePlacement(placement) : null;
  var variationPlacement = placement ? getVariationPlacement(placement) : null;
  var scrollTop = scroll.scrollTop,
      scrollLeft = scroll.scrollLeft;
  var offsets;

  switch (basePlacement) {
    case top:
      offsets = {
        x: reference.x + reference.width / 2 - element.width / 2 - scrollLeft,
        y: reference.y - element.height - scrollTop
      };
      break;

    case bottom:
      offsets = {
        x: reference.x + reference.width / 2 - element.width / 2 - scrollLeft,
        y: reference.y + reference.height - scrollTop
      };
      break;

    case right:
      offsets = {
        x: reference.x + reference.width - scrollLeft,
        y: reference.y + reference.height / 2 - element.height / 2 - scrollTop
      };
      break;

    case left:
      offsets = {
        x: reference.x - element.width - scrollLeft,
        y: reference.y + reference.height / 2 - element.height / 2 - scrollTop
      };
      break;

    default:
      offsets = {
        x: reference.x - scrollLeft,
        y: reference.y - scrollTop
      };
  }

  var mainAxis = basePlacement ? getMainAxisFromPlacement(basePlacement) : null;

  if (mainAxis != null) {
    var len = mainAxis === 'y' ? 'height' : 'width';

    switch (variationPlacement) {
      case start:
        offsets[mainAxis] = Math.floor(offsets[mainAxis]) - Math.floor(reference[len] / 2 - element[len] / 2);
        break;

      case end:
        offsets[mainAxis] = Math.floor(offsets[mainAxis]) + Math.ceil(reference[len] / 2 - element[len] / 2);
        break;
    }
  }

  return offsets;
});

function getWindowScroll(node) {
  var win = getWindow(node);
  var scrollLeft = win.pageXOffset;
  var scrollTop = win.pageYOffset;
  return {
    scrollLeft: scrollLeft,
    scrollTop: scrollTop
  };
}

function getHTMLElementScroll(element) {
  return {
    scrollLeft: element.scrollLeft,
    scrollTop: element.scrollTop
  };
}

function getElementScroll(node) {
  if (node === getWindow(node) || !(node instanceof HTMLElement)) {
    return getWindowScroll(node);
  } else {
    return getHTMLElementScroll(node);
  }
}

function getOffsetParent(element) {
  var offsetParent = element instanceof HTMLElement ? element.offsetParent : null;
  var window = getWindow(element);

  if (getNodeName(offsetParent) === 'BODY') {
    return window;
  }

  return offsetParent || window;
}

var sumScroll = function sumScroll(scrollParents) {
  return scrollParents.reduce(function (scroll, scrollParent) {
    var nodeScroll = getElementScroll(scrollParent);
    scroll.scrollTop += nodeScroll.scrollTop;
    scroll.scrollLeft += nodeScroll.scrollLeft;
    return scroll;
  }, {
    scrollTop: 0,
    scrollLeft: 0
  });
};

function getCommonTotalScroll(reference, referenceScrollParents, popperScrollParents, limiter) {
  var offsetParent = getOffsetParent(reference);
  var index = referenceScrollParents.findIndex(function (node) {
    return node === (limiter || offsetParent);
  });
  var scrollParents = referenceScrollParents.slice(0, index === -1 ? undefined : index);
  return sumScroll(scrollParents);
}

function popperOffsets(state) {
  // Offsets are the actual position the popper needs to have to be
  // properly positioned near its reference element
  // This is the most basic placement, and will be adjusted by
  // the modifiers in the next step
  state.modifiersData.popperOffsets = computeOffsets({
    reference: state.measures.reference,
    element: state.measures.popper,
    strategy: 'absolute',
    placement: state.placement,
    scroll: getCommonTotalScroll(state.elements.reference, state.scrollParents.reference, state.scrollParents.popper)
  });
  return state;
}
var popperOffsets$1 = {
  name: 'popperOffsets',
  enabled: true,
  phase: 'read',
  fn: popperOffsets,
  data: {}
};

var getBoundingClientRect = (function (element) {
  return JSON.parse(JSON.stringify(element.getBoundingClientRect()));
});

var getDocumentElement = (function (element) {
  return element.ownerDocument.documentElement;
});

// clipping (or hiding) overflowing elements with a position different from `initial`

function getClippingParent(element, offsetElement) {
  var scrollParent = getScrollParent(element);
  var offsetParent = offsetElement || getOffsetParent(element);
  var win = getWindow(element);
  return offsetParent === win ? getDocumentElement(element) : scrollParent.contains(offsetParent) ? scrollParent : getClippingParent(getScrollParent(getParentNode(scrollParent)), offsetParent);
}

var getDocumentRect = (function (element) {
  var win = getWindow(element);
  var documentRect = getElementClientRect(getDocumentElement(element));
  documentRect.height = Math.max(documentRect.height, win.innerHeight);
  documentRect.width = Math.max(documentRect.width, win.innerWidth);
  return documentRect;
});

var rectToClientRect = (function (rect) {
  return Object.assign({}, rect, {
    left: rect.x,
    top: rect.y,
    right: rect.x + rect.width,
    bottom: rect.y + rect.height
  });
});

function detectOverflow(state, options) {
  if (options === void 0) {
    options = {
      boundaryElement: getClippingParent(state.elements.popper)
    };
  }

  var popperElement = state.elements.popper;
  var referenceElement = state.elements.reference;
  var popperRect = state.measures.popper;
  var documentElement = getDocumentElement(options.boundaryElement);

  if (!options.boundaryElement.contains(popperElement)) {
    if (process.env.NODE_ENV !== "production") {
      console.error('Popper: "detectOverflow" can accept as `boundaryElement` only a parent node of the provided popper.');
    }

    return state;
  }

  var boundaryClientRect = documentElement === options.boundaryElement ? rectToClientRect(getDocumentRect(documentElement)) : getBoundingClientRect(options.boundaryElement);
  var referenceClientRect = getBoundingClientRect(referenceElement);
  var popperOffsets = computeOffsets({
    reference: referenceClientRect,
    element: popperRect,
    strategy: 'absolute',
    placement: state.placement,
    scroll: {
      scrollTop: 0,
      scrollLeft: 0
    }
  });
  var popperClientRect = rectToClientRect(Object.assign({}, popperRect, {}, popperOffsets));
  state.modifiersData.detectOverflow = {
    top: boundaryClientRect.top - popperClientRect.top,
    bottom: popperClientRect.bottom - boundaryClientRect.bottom,
    left: boundaryClientRect.left - popperClientRect.left,
    right: popperClientRect.right - boundaryClientRect.right
  };
  return state;
}
var detectOverflow$1 = {
  name: 'detectOverflow',
  enabled: true,
  phase: 'read',
  fn: detectOverflow,
  requires: ['popperOffsets'],
  data: {}
};

// This modifier takes the Popper state and prepares some StyleSheet properties
// that can be applied to the popper element to make it render in the expected position.
var mapStrategyToPosition = function mapStrategyToPosition(strategy) {
  switch (strategy) {
    case 'fixed':
      return 'fixed';

    case 'absolute':
    default:
      return 'absolute';
  }
};
var computePopperStyles = function computePopperStyles(_ref) {
  var offsets = _ref.offsets,
      strategy = _ref.strategy,
      gpuAcceleration = _ref.gpuAcceleration;

  // @1x displays in Chrome can be blurry due to non-rounded offsets, but this
  // introduces slight positioning errors. TODO: somehow solve this better
  // by default it is active, disable it only if explicitly set to false
  if (gpuAcceleration === false || window.devicePixelRatio < 2) {
    return {
      top: offsets.y + "px",
      left: offsets.x + "px",
      position: mapStrategyToPosition(strategy)
    };
  } else {
    return {
      top: '0px',
      left: '0px',
      transform: "translate3d(" + offsets.x + "px, " + offsets.y + "px, 0)",
      position: mapStrategyToPosition(strategy)
    };
  }
};
var computeArrowStyles = function computeArrowStyles(_ref2) {
  var offsets = _ref2.offsets,
      gpuAcceleration = _ref2.gpuAcceleration;

  if (gpuAcceleration) {
    return {
      top: offsets.y + "px",
      left: offsets.x + "px",
      position: 'absolute'
    };
  } else {
    return {
      transform: "translate3d(" + offsets.x + "px, " + offsets.y + "px, 0)",
      position: 'absolute'
    };
  }
};
function computeStyles(state, options) {
  if (options === void 0) {
    options = {};
  }

  var _options = options,
      _options$gpuAccelerat = _options.gpuAcceleration,
      gpuAcceleration = _options$gpuAccelerat === void 0 ? true : _options$gpuAccelerat;
  state.modifiersData.computeStyles = {
    styles: {
      // popper offsets are always available
      popper: computePopperStyles({
        offsets: state.modifiersData.popperOffsets,
        strategy: state.options.strategy,
        gpuAcceleration: gpuAcceleration
      }),
      // arrow offsets may not be available
      arrow: state.modifiersData.arrow != null ? computeArrowStyles({
        offsets: state.modifiersData.arrow,
        gpuAcceleration: gpuAcceleration
      }) : undefined
    },
    attributes: {
      popper: {
        'data-popper-placement': state.placement
      }
    }
  };
  return state;
}
var computeStyles$1 = {
  name: 'computeStyles',
  enabled: true,
  phase: 'afterMain',
  fn: computeStyles,
  data: {}
};

// This modifier takes the styles prepared by the `computeStyles` modifier
// and applies them to the HTMLElements such as popper and arrow
function applyStyles(state) {
  var data = state.modifiersData.computeStyles;
  Object.keys(state.elements).forEach(function (name) {
    var style = data.styles[name] || {};
    var attributes = data.attributes[name] || {};
    var element = state.elements[name]; // arrow is optional

    if (!element) {
      return;
    }

    Object.assign(element.style, style);
    Object.entries(attributes).forEach(function (args) {
      return element.setAttribute.apply(element, args);
    });
  });
  return state;
}
function onDestroy(state) {
  var data = state.modifiersData.computeStyles;
  Object.keys(state.elements).forEach(function (name) {
    var element = state.elements[name];
    var styleProperties = Object.keys(data.styles.hasOwnProperty(name) ? Object.assign({}, data.styles[name]) : {});
    var attributes = data.attributes[name] || {}; // Set all values to an empty string to unset them

    var style = styleProperties.reduce(function (style, property) {
      var _Object$assign;

      return Object.assign({}, style, (_Object$assign = {}, _Object$assign[String(property)] = '', _Object$assign));
    }, {}); // arrow is optional

    if (!element) {
      return;
    } // Flow doesn't support to extend this property, but it's the most
    // effective way to apply styles to an HTMLElement
    // $FlowIgnore


    Object.assign(element.style, style);
    Object.keys(attributes).forEach(function (attribute) {
      return element.removeAttribute(attribute);
    });
  });
}
var applyStyles$1 = {
  name: 'applyStyles',
  enabled: true,
  phase: 'write',
  fn: applyStyles,
  onDestroy: onDestroy,
  requires: ['computeStyles']
};

function distanceAndSkiddingToXY(placement, measures, getOffsets) {
  var basePlacement = getBasePlacement(placement);
  var invertDistance = ['left', 'top'].includes(basePlacement) ? -1 : 1;
  var invertSkidding = ['top', 'bottom'].includes(basePlacement) ? -1 : 1;

  var _getOffsets = getOffsets(Object.assign({}, measures, {
    placement: placement
  })),
      distance = _getOffsets[0],
      skidding = _getOffsets[1];

  distance = (distance || 0) * invertDistance;
  skidding = (distance || 0) * invertSkidding;
  return ['left', 'right'].includes(basePlacement) ? [distance, skidding] : [skidding, distance];
}
function offset(state, options) {
  if (options && typeof options.offset === 'function') {
    var _distanceAndSkiddingT = distanceAndSkiddingToXY(state.placement, state.measures, options.offset),
        x = _distanceAndSkiddingT[0],
        y = _distanceAndSkiddingT[1];

    state.modifiersData.popperOffsets.x += x;
    state.modifiersData.popperOffsets.y += y;
  }

  return state;
}
var offset$1 = {
  name: 'offset',
  enabled: true,
  phase: 'main',
  fn: offset
};

var hash = {
  left: 'right',
  right: 'left',
  bottom: 'top',
  top: 'bottom'
};
var getOppositePlacement = (function (placement) {
  return placement.replace(/left|right|bottom|top/g, function (matched) {
    return hash[matched];
  });
});

var mergePaddingObject = (function (paddingObject) {
  return Object.assign({
    top: 0,
    bottom: 0,
    left: 0,
    right: 0
  }, paddingObject);
});

function flip(state, options) {
  if (options === void 0) {
    options = {};
  }

  var placement = state.placement;
  var defaultBehavior = [state.options.placement, getOppositePlacement(state.options.placement)];
  var _options = options,
      _options$behavior = _options.behavior,
      behavior = _options$behavior === void 0 ? defaultBehavior : _options$behavior,
      _options$padding = _options.padding,
      padding = _options$padding === void 0 ? 0 : _options$padding;
  var overflow = state.modifiersData.detectOverflow;
  var flipIndex = state.modifiersData.flip.index;
  var paddingObject = mergePaddingObject(typeof padding !== 'number' ? padding : expandToHashMap(padding, basePlacements));
  var flippedPlacement = behavior[flipIndex];

  if (!flippedPlacement && placement !== state.options.placement) {
    state.placement = state.options.placement;
    state.reset = true;
    return state;
  }

  if (!flippedPlacement && placement === state.options.placement) {
    return state;
  }

  var basePlacement = getBasePlacement(flippedPlacement);
  var fits = overflow[basePlacement] + paddingObject[basePlacement] <= 0;

  if (!fits) {
    state.modifiersData.flip.index += 1;
    state.reset = true;
    return state;
  } else if (fits && state.placement !== flippedPlacement) {
    state.placement = flippedPlacement;
    state.reset = true;
    return state;
  }

  return state;
}
var flip$1 = {
  name: 'flip',
  enabled: true,
  phase: 'main',
  fn: flip,
  requires: ['detectOverflow'],
  data: {
    index: 0
  }
};

var getAltAxis = (function (axis) {
  return axis === 'x' ? 'y' : 'x';
});

var within = (function (min, value, max) {
  return Math.max(min, Math.min(value, max));
});

function preventOverflow(state, options) {
  if (options === void 0) {
    options = {};
  }

  var _options = options,
      _options$mainAxis = _options.mainAxis,
      checkMainAxis = _options$mainAxis === void 0 ? true : _options$mainAxis,
      _options$altAxis = _options.altAxis,
      checkAltAxis = _options$altAxis === void 0 ? false : _options$altAxis,
      _options$tether = _options.tether,
      tether = _options$tether === void 0 ? true : _options$tether,
      _options$padding = _options.padding,
      padding = _options$padding === void 0 ? 0 : _options$padding;
  var overflow = state.modifiersData.detectOverflow;
  var basePlacement = getBasePlacement(state.placement);
  var mainAxis = getMainAxisFromPlacement(basePlacement);
  var altAxis = getAltAxis(mainAxis);
  var popperOffsets = state.modifiersData.popperOffsets;
  var referenceRect = state.measures.reference;
  var popperRect = state.measures.popper;
  var paddingObject = mergePaddingObject(typeof padding !== 'number' ? padding : expandToHashMap(padding, basePlacements));

  if (checkMainAxis) {
    var mainSide = mainAxis === 'y' ? top : left;
    var altSide = mainAxis === 'y' ? bottom : right;
    var len = mainAxis === 'y' ? 'height' : 'width';
    var offset = popperOffsets[mainAxis];
    var min = popperOffsets[mainAxis] + overflow[mainSide] + paddingObject[mainSide];
    var max = popperOffsets[mainAxis] - overflow[altSide] - paddingObject[altSide];
    var tetherMin = state.modifiersData.popperOffsets[mainAxis] - referenceRect[len] / 2 + popperRect[len] / 2;
    var tetherMax = state.modifiersData.popperOffsets[mainAxis] + referenceRect[len] / 2 - popperRect[len] / 2;
    var isReferenceLarger = referenceRect[len] > popperRect[len];
    state.modifiersData.popperOffsets[mainAxis] = within(tether ? Math.min(min, isReferenceLarger ? tetherMax : tetherMin) : min, offset, tether ? Math.max(max, isReferenceLarger ? tetherMin : tetherMax) : max);
  }

  if (checkAltAxis) {
    var _mainSide = mainAxis === 'x' ? top : left;

    var _altSide = mainAxis === 'x' ? bottom : right;

    state.modifiersData.popperOffsets[altAxis] = within(popperOffsets[altAxis] + overflow[_mainSide] + paddingObject[_mainSide], popperOffsets[altAxis], popperOffsets[altAxis] - overflow[_altSide] - paddingObject[_altSide]);
  }

  return state;
}
var preventOverflow$1 = {
  name: 'preventOverflow',
  enabled: true,
  phase: 'main',
  fn: preventOverflow,
  requires: ['detectOverflow']
};

function arrow(state, options) {
  var _state$modifiersData$;

  if (options === void 0) {
    options = {};
  }

  var _options = options,
      _options$element = _options.element,
      arrowElement = _options$element === void 0 ? '[data-popper-arrow]' : _options$element; // CSS selector

  if (typeof arrowElement === 'string') {
    arrowElement = state.elements.popper.querySelector(arrowElement);

    if (!arrowElement) {
      return state;
    }
  }

  if (!state.elements.popper.contains(arrowElement)) {
    if (process.env.NODE_ENV !== "production") {
      console.error('Popper: "arrow" modifier\'s `element` must be a child of the popper element.');
    }

    return state;
  }

  state.elements.arrow = arrowElement;
  var popperOffsets = state.modifiersData.popperOffsets;
  var basePlacement = getBasePlacement(state.placement);
  var axis = getMainAxisFromPlacement(basePlacement);
  var isVertical = ['left', 'right'].includes(basePlacement);
  var len = isVertical ? 'height' : 'width';
  var arrowElementRect = addClientRectMargins(getElementClientRect(arrowElement), arrowElement);
  var commonTotalScroll = getCommonTotalScroll(state.elements.reference, state.scrollParents.reference, state.scrollParents.popper);
  var endDiff = state.measures.reference[len] + state.measures.reference[axis] - popperOffsets[axis] - state.measures.popper[len];
  var startDiff = popperOffsets[axis] - state.measures.reference[axis]; // Only center to reference if the popper is longer than reference along the
  // axis

  var centerToReference = state.measures.popper[len] > state.measures.reference[len] ? endDiff / 2 - startDiff / 2 - commonTotalScroll[axis === 'y' ? 'scrollTop' : 'scrollLeft'] : 0;
  var center = state.measures.popper[len] / 2 - arrowElementRect[len] / 2 + centerToReference; // Make sure the arrow doesn't overflow the popper if the center point is
  // outside of the popper bounds

  center = within(0, center, state.measures.popper[len] - arrowElementRect[len]); // Prevents breaking syntax highlighting...

  var axisProp = axis;
  state.modifiersData.arrow = (_state$modifiersData$ = {}, _state$modifiersData$[axisProp] = center, _state$modifiersData$);
  return state;
}
var arrow$1 = {
  name: 'arrow',
  enabled: true,
  phase: 'main',
  fn: arrow,
  requires: ['popperOffsets']
};



var modifiers = /*#__PURE__*/Object.freeze({
  __proto__: null,
  popperOffsets: popperOffsets$1,
  detectOverflow: detectOverflow$1,
  computeStyles: computeStyles$1,
  applyStyles: applyStyles$1,
  offset: offset$1,
  flip: flip$1,
  preventOverflow: preventOverflow$1,
  arrow: arrow$1
});

var defaultModifiers = Object.values(modifiers);
Popper.defaultModifiers = defaultModifiers;

module.exports = Popper;
//# sourceMappingURL=popper.js.map
