// DOM Utils
import getElementClientRect from './dom-utils/getElementClientRect';
import listScrollParents from './dom-utils/listScrollParents';
import getWindow from './dom-utils/getWindow';
import addClientRectMargins from './dom-utils/addClientRectMargins'; // Pure Utils

import unwrapJqueryElement from './utils/unwrapJqueryElement';
import orderModifiers from './utils/orderModifiers';
import expandToHashMap from './utils/expandToHashMap';
import debounce from './utils/debounce';
import validateModifiers from './utils/validateModifiers';
var INVALID_ELEMENT_ERROR = 'Popper: Invalid reference or popper argument provided to Popper, they must be either a valid DOM element or a jQuery-wrapped DOM element.';
var INFINITE_LOOP_ERROR = 'Popper: An infinite loop in the modifiers cycle has been detected! The cycle has been interrupted to prevent a browser crash.';

var areValidElements = function areValidElements() {
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  return !args.some(function (element) {
    return !(element instanceof Element);
  });
};

var defaultOptions = {
  placement: 'bottom',
  eventListeners: {
    scroll: true,
    resize: true
  },
  modifiers: [],
  strategy: 'absolute'
};

var Popper =
/*#__PURE__*/
function () {
  function Popper(reference, popper, options) {
    var _this = this;

    if (options === void 0) {
      options = defaultOptions;
    }

    this.state = {
      placement: 'bottom',
      orderedModifiers: [],
      options: defaultOptions,
      modifiersData: {}
    };
    this.update = debounce(function () {
      return new Promise(function (success) {
        _this.forceUpdate();

        success(_this.state);
      });
    });
    // Unwrap `reference` and `popper` elements in case they are
    // wrapped by jQuery, otherwise consume them as is
    var referenceElement = unwrapJqueryElement(reference);
    var popperElement = unwrapJqueryElement(popper);
    this.state.elements = {
      reference: referenceElement,
      popper: popperElement
    }; // Store options into state

    this.state.options = Object.assign({}, defaultOptions, {}, options); // Don't proceed if `reference` or `popper` are invalid elements

    if (!areValidElements(referenceElement, popperElement)) {
      if (process.env.NODE_ENV !== "production") {
        console.error(INVALID_ELEMENT_ERROR);
      }

      return;
    }

    this.state.scrollParents = {
      reference: listScrollParents(referenceElement),
      popper: listScrollParents(popperElement)
    }; // Validate the provided modifiers so that the consumer will get warned
    // of one of the custom modifiers is invalid for any reason

    if (process.env.NODE_ENV !== "production") {
      validateModifiers(this.state.options.modifiers);
    } // Order `options.modifiers` so that the dependencies are fulfilled
    // once the modifiers are executed


    this.state.orderedModifiers = orderModifiers([].concat(this.constructor.defaultModifiers, this.state.options.modifiers)) // Apply user defined preferences to modifiers
    .map(function (modifier) {
      return Object.assign({}, modifier, {}, _this.state.options.modifiers.find(function (_ref) {
        var name = _ref.name;
        return name === modifier.name;
      }));
    }); // Modifiers have the opportunity to execute some arbitrary code before
    // the first update cycle is ran, the order of execution will be the same
    // defined by the modifier dependencies directive.
    // The `onLoad` function may add or alter the options of themselves

    this.state.orderedModifiers.forEach(function (_ref2) {
      var onLoad = _ref2.onLoad,
          enabled = _ref2.enabled;
      return enabled && onLoad && onLoad(_this.state);
    });
    this.update().then(function () {
      // After the first update completed, enable the event listeners
      _this.enableEventListeners(_this.state.options.eventListeners);
    });
  } // Async and optimistically optimized update
  // it will not be executed if not necessary
  // debounced, so that it only runs at most once-per-tick


  var _proto = Popper.prototype;

  // Syncronous and forcefully executed update
  // it will always be executed even if not necessary, usually NOT needed
  // use Popper#update instead
  _proto.forceUpdate = function forceUpdate() {
    var _this2 = this;

    var _this$state$elements = this.state.elements,
        referenceElement = _this$state$elements.reference,
        popperElement = _this$state$elements.popper; // Don't proceed if `reference` or `popper` are not valid elements anymore

    if (!areValidElements(referenceElement, popperElement)) {
      if (process.env.NODE_ENV !== "production") {
        console.error(INVALID_ELEMENT_ERROR);
      }

      return;
    } // Get initial measurements
    // these are going to be used to compute the initial popper offsets
    // and as cache for any modifier that needs them later


    this.state.measures = {
      reference: getElementClientRect(referenceElement),
      // CSS marginsc an be applied to popper elements to quickly
      // apply offsets dynamically based on some CSS selectors.
      // For this reason we include margins in this calculation.
      popper: addClientRectMargins(getElementClientRect(popperElement), popperElement)
    }; // Modifiers have the ability to read the current Popper state, included
    // the popper offsets, and modify it to address specifc cases

    this.state.reset = false; // Cache the placement in cache to make it available to the modifiers
    // modifiers will modify this one (rather than the one in options)

    this.state.placement = this.state.options.placement;
    this.state.orderedModifiers.forEach(function (modifier) {
      return _this2.state.modifiersData[modifier.name] = Object.assign({}, modifier.data);
    });
    var __debug_loops__ = 0;

    for (var index = 0; index < this.state.orderedModifiers.length; index++) {
      if (process.env.NODE_ENV !== "production") {
        __debug_loops__ += 1;

        if (__debug_loops__ > 100) {
          console.error(INFINITE_LOOP_ERROR);
          break;
        }
      }

      if (this.state.reset === true) {
        this.state.reset = false;
        index = -1;
        continue;
      }

      var _this$state$orderedMo = this.state.orderedModifiers[index],
          fn = _this$state$orderedMo.fn,
          enabled = _this$state$orderedMo.enabled,
          options = _this$state$orderedMo.options;

      if (enabled && typeof fn === 'function') {
        this.state = fn(this.state, options);
      }
    }
  };

  _proto.enableEventListeners = function enableEventListeners(eventListeners) {
    var _this3 = this;

    var _ref3 = typeof eventListeners === 'boolean' ? expandToHashMap(eventListeners, ['scroll', 'resize']) : eventListeners,
        scroll = _ref3.scroll,
        resize = _ref3.resize;

    if (scroll) {
      var scrollParents = [].concat(this.state.scrollParents.reference, this.state.scrollParents.popper);
      scrollParents.forEach(function (scrollParent) {
        return scrollParent.addEventListener('scroll', _this3.update, {
          passive: true
        });
      });
    }

    if (resize) {
      var window = getWindow(this.state.elements.popper);
      window.addEventListener('resize', this.update, {
        passive: true
      });
    }
  };

  _proto.destroy = function destroy() {
    var _this4 = this;

    // Remove scroll event listeners
    var scrollParents = [].concat(this.state.scrollParents.reference, this.state.scrollParents.popper);
    scrollParents.forEach(function (scrollParent) {
      return scrollParent.removeEventListener('scroll', _this4.update);
    }); // Remove resize event listeners

    var window = getWindow(this.state.elements.popper);
    window.removeEventListener('resize', this.update); // Run `onDestroy` modifier methods

    this.state.orderedModifiers.forEach(function (_ref4) {
      var onDestroy = _ref4.onDestroy,
          enabled = _ref4.enabled;
      return enabled && onDestroy && onDestroy(_this4.state);
    });
  };

  return Popper;
}();

export { Popper as default };
Popper.defaultModifiers = [];